/***********************************************************************
 * A SystemVerilog testbench for an instruction register.
 * The course labs will convert this to an object-oriented testbench
 * with constrained random test generation, functional coverage, and
 * a scoreboard for self-verification.
 *
 **********************************************************************/

module instr_register_test
  import instr_register_pkg::*;  // user-defined types are defined in instr_register_pkg.sv
  (input  logic          clk,
   output logic          load_en,
   output logic          reset_n,
   output operand_t      operand_a,
   output operand_t      operand_b,
   output opcode_t       opcode,
   output address_t      write_pointer,
   output address_t      read_pointer,
   input  instruction_t  instruction_word
  );

  timeunit 1ns; timeprecision 1ns;

  initial begin
    $display("\nReseting the instruction register...");
    write_pointer  = 5'h00;         // initialize write pointer
    read_pointer   = 5'h1F;         // initialize read pointer
    load_en        = 1'b0;          // initialize load control line
    reset_n        = 1'b0;          // assert reset_n (active low) - should be non-blocking
    repeat (2) @(posedge clk) ;     // hold in reset for 2 clock cycles
    reset_n        = 1'b1;          // deassert reset_n (active low)

    $display("\nWriting values to register stack...");
    @(posedge clk)
    #1 load_en = 1'b1;  // enable writing to register
    repeat (3) begin
      @(posedge clk) 
      #1 randomize_transaction();
      #8 print_transaction();
    end
    @(posedge clk) 
    #1 load_en = 1'b0;  // turn-off writing to register

    // read back and display same three register locations
    $display("\nReading back the same register locations written...");
    for (int i=0; i<3; i++) begin
      // Later labs will replace this loop with iterating through a
      // scoreboard to determine which addresses were written and
      // the expected values to be read back
      @(posedge clk);
      #1 read_pointer = i;
      #8 print_results();
    end

    @(posedge clk) $finish();
  end

  function void randomize_transaction();
    // A later lab will replace this function with SystemVerilog
    // constrained random values
    //
    // The static temp variable is required in order to write to fixed
    // addresses of 0, 1 and 2.  This will be replaced with randomized
    // write_pointer values in a later lab
    //
    static int temp = 0;
    operand_a     = $urandom_range(0, 30) - 15;       // between -15 and 15
    operand_b     = $urandom_range(0, 15);            // between 0 and 15
    opcode        = opcode_t'($urandom_range(0, 7));  // between 0 and 7, cast to opcode_t type
    write_pointer = temp++;
  endfunction: randomize_transaction
  
  function void print_transaction();
    $display("Write to register location %2d: opcode = %0d (%5s) operand_a =%3d  operand_b =%3d", 
 	     write_pointer, opcode, opcode.name(), operand_a, operand_b);
  endfunction

  function void print_results();
    $display("Read from register location %2d: opcode = %0d (%5s) operand_a =%3d  operand_b =%3d", 
	     read_pointer, instruction_word.opc, instruction_word.opc.name(), instruction_word.op_a, instruction_word.op_b);
  endfunction

endmodule: instr_register_test
